// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package controllers

import (
	"fmt"
	"net/http"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/middleware"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/repos"
	"gitlab.bit9.local/psc/ems/hinting/internal/dbwrapper"
	"gitlab.bit9.local/psc/ems/hinting/internal/redisclient"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"go.uber.org/zap"

	"github.com/gin-gonic/gin"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

// HintController -- Structure used to handle the generation of hints
type HintController struct {
	generator       hints.IHintGenerator
	deviceRepo      repos.Repository[*models.Device]
	deviceGroupRepo repos.Repository[*models.DeviceGroup]
	orgRepo         repos.Repository[*models.Organization]
	envRepo         repos.Repository[*models.Environment]
	auditorFactory  hints.AuditorFactory
}

//goland:noinspection GoMixedReceiverTypes
func (hc *HintController) withGenerator(generator hints.IHintGenerator) *HintController {
	hc.generator = generator
	return hc
}

// NewHintController -- Instantiate a new HintController
func NewHintController(deviceRepo repos.Repository[*models.Device], deviceGroupRepo repos.Repository[*models.DeviceGroup],
	orgRepo repos.Repository[*models.Organization], globalRepo repos.Repository[*models.Environment],
) HintController {
	return NewHintControllerWithAuditor(deviceRepo, deviceGroupRepo, orgRepo, globalRepo, hints.NewHintAuditorFromCfg)
}

// NewHintControllerWithAuditor -- Instantiate a new HintController along with a hint audit factory function
func NewHintControllerWithAuditor(deviceRepo repos.Repository[*models.Device], deviceGroupRepo repos.Repository[*models.DeviceGroup],
	orgRepo repos.Repository[*models.Organization], globalRepo repos.Repository[*models.Environment], auditorFactory hints.AuditorFactory,
) HintController {
	return HintController{
		generator:       hints.NewHintGenerator(),
		deviceRepo:      deviceRepo,
		deviceGroupRepo: deviceGroupRepo,
		orgRepo:         orgRepo,
		envRepo:         globalRepo,
		auditorFactory:  auditorFactory,
	}
}

// NewHintControllerFromDbWithCache -- Instantiate a new HintController from the existing redis cache
//   - no hint auditing factory will be added!
func NewHintControllerFromDbWithCache(db *dbwrapper.DbClient, redisClient redisclient.RedisClient) HintController {
	return NewHintController(
		repos.NewCachedRepo[*models.Device](redisClient, db),
		repos.NewCachedRepo[*models.DeviceGroup](redisClient, db),
		repos.NewCachedRepo[*models.Organization](redisClient, db),
		repos.NewCachedRepo[*models.Environment](redisClient, db),
	)
}

//goland:noinspection GoMixedReceiverTypes
func (hc HintController) GetHints(c *gin.Context) {
	var logFields []zap.Field = make([]zap.Field, 0)

	defer func() {
		c.Set(string(middleware.ExtraLogFields), logFields)
	}()

	var inputRequestMessage hints.HintRequestMessage
	err := c.ShouldBindJSON(&inputRequestMessage)
	if err != nil {
		_ = c.AbortWithError(http.StatusBadRequest, fmt.Errorf("failed to parse request body %w", err))
		return
	}

	logFields = append(logFields, zap.Object("input_request_message", inputRequestMessage))

	orgKeyFromLib, err := orgkeylib.Decode(inputRequestMessage.OrgKey)
	if err != nil {
		_ = c.AbortWithError(http.StatusBadRequest, fmt.Errorf("org key failed to decode: %w", err))
		return
	}
	logFields = append(logFields, zap.String("env_name", orgKeyFromLib.String()))

	device := &models.Device{Id: inputRequestMessage.DeviceId, OrgKey: inputRequestMessage.OrgKey}

	if err = hc.deviceRepo.Get(c.Request.Context(), device); err != nil {
		_ = c.AbortWithError(http.StatusInternalServerError, err)
		return
	}
	device.DeviceType = &inputRequestMessage.DeviceType
	device.DeploymentType = &inputRequestMessage.DeploymentType

	logFields = append(logFields, zap.Object("device_info", device))

	group := &models.DeviceGroup{Id: inputRequestMessage.DeviceGroupId, OrgKey: inputRequestMessage.OrgKey}
	if err = hc.deviceGroupRepo.Get(c.Request.Context(), group); err != nil {
		_ = c.AbortWithError(http.StatusInternalServerError, err)
		return
	}

	logFields = append(logFields, zap.Object("group_info", group))

	org := &models.Organization{OrgKey: inputRequestMessage.OrgKey}
	if err = hc.orgRepo.Get(c.Request.Context(), org); err != nil {
		_ = c.AbortWithError(http.StatusInternalServerError, err)
		return
	}
	logFields = append(logFields, zap.Object("org_info", org))

	var env *models.Environment
	if orgKeyFromLib != nil {
		env = &models.Environment{Name: orgKeyFromLib.Env()}
		if err = hc.envRepo.Get(c.Request.Context(), env); err != nil {
			_ = c.AbortWithError(http.StatusInternalServerError, err)
			return
		}
		logFields = append(logFields, zap.Object("env_info", env))
	}

	inAuditMode := inputRequestMessage.AuditHintResult != nil
	hintRequest := hints.NewHintRequestFromBase(env, org, group, device, inputRequestMessage.HintRequestBase)
	response := hc.generator.GenerateHints(hintRequest, inAuditMode)
	c.JSON(http.StatusOK, response)

	// Handle auditing checks
	if inAuditMode {
		hc.auditorFactory(&inputRequestMessage, response).Audit(c)
	}
}
