// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

// Package devicechangetype -- Contains enumeration values for device change types
package devicechangetype

import "golang.org/x/exp/maps"

//go:generate stringer -type=DeviceGroupChangeType

// DeviceChangeType -- Type alias of allowed device change types
type DeviceChangeType int

// NOTE: Changes here should also be reflected in `test/functional/suites/lib/models/enumerations.py`

// NOTE: Keep in sync with:
//   - https://gitlab.bit9.local/cb-defense/cbd/-/blob/main/backend/domainModel/src/main/java/com/scargo/domainmodel/entity/support/ChangeTimestampType.java
//
//goland:noinspection GoSnakeCaseUsage
const (
	DEV_GROUP_CHANGE                   DeviceChangeType = 301 // Device moved to new group
	DEV_HASH_SAMPLE_REQUEST            DeviceChangeType = 302 // Hash sample requested from sensor
	DEV_SENSOR_LOGS_REQUEST            DeviceChangeType = 303 // Sensor logs requested
	DEV_HASH_DELETE_REQUEST            DeviceChangeType = 304 // Request to delete a hash on the sensor
	DEV_SENSOR_TESTKIT_UPDATE          DeviceChangeType = 305 // Request to update the sensor to a test kit
	DEV_SENSOR_CERTIFICATE_REQUEST     DeviceChangeType = 306 // Request to upload sensor certificate
	DEV_SENSOR_CONFIG_UPDATE           DeviceChangeType = 307 // Request to update the sensor config for reset and mode
	DEV_SENSOR_TRIGGER_PERFMON         DeviceChangeType = 308 // Request to trigger performance monitoring for a device
	DEV_SENSOR_KIT_UPDATE              DeviceChangeType = 309 // Request to update the sensor to kit version specified by {@linkplain Device.TARGET_SENSOR_VERSION} property
	DEV_SENSOR_ACTION                  DeviceChangeType = 310 // Request to trigger sensor actions (latest)
	DEV_LAST_SENSOR_ACTION             DeviceChangeType = 311 // Keep track of last sensor action time
	DEV_LAST_HASH_DELETE_SENSOR_ACTION DeviceChangeType = 312 // Last hash delete sensor action time
	DEV_LAST_HASH_UPLOAD_SENSOR_ACTION DeviceChangeType = 313 // Last hash upload sensor action time
	DEV_LAST_CONFIG_SENSOR_ACTION      DeviceChangeType = 314 // Last config sensor action time
	DEV_DASHBOARD_DEREGISTER           DeviceChangeType = 315 // Deregister time stamp
	DEV_ASYNC_REPUTATION               DeviceChangeType = 316 // Deprecated; Async reputation time stamp (UNUSED DUE TO PERFORMANCE IMPACT)
	DEV_LIVE_QUERY_CHANGE              DeviceChangeType = 317 // Last live query update time
	DEV_SIG_PACK_BEFORE_REBOOT         DeviceChangeType = 318 // time when backend sent a policy hint down to the sensor in response to a device status message when the sensor has not rebooted since aug 7, 2019
	DEV_SIG_PACK_AFTER_REBOOT          DeviceChangeType = 319 // time when backend sent a policy hint down to the sensor in response to a device status message when the sensor has rebooted since aug 7, 2019.
	DEV_SIG_PACK_PATCH                 DeviceChangeType = 320 // time when backend sent a policy hint down to the sensor in response to a device status message when the sensor is already on the keyed sig pack.
	DEV_REREGISTER                     DeviceChangeType = 321 // time when a device was set to reregister itself. No updates will be considered for this device now, as it was incorrectly cloned
)

/**
 * DEV_SIG_PACK_BEFORE_REBOOT, DEV_SIG_PACK_AFTER_REBOOT, DEV_SIG_PACK_PATCH
 *
 * DSER-17722: After receiving a policy update, sensor immediately sends a device status message,
 * which results in an infinite loop on the sensor asking for policy updates and sending device
 * status message. Purpose of saving these timestamps below is to ensure that backend sends policy
 * update hint only once to the sensor and so it does not enter this infinite loop.
 */

//goland:noinspection GoSnakeCaseUsage,GoUnusedGlobalVariable
var (
	// TranslateDeviceChangeTypeString -- Translation from string -> DeviceChangeType
	//  - limited to those device change types used by hinting
	TranslateDeviceChangeTypeString = map[string]DeviceChangeType{ //nolint:gochecknoglobals // we don't care
		DEV_GROUP_CHANGE.String():               DEV_GROUP_CHANGE,
		DEV_HASH_DELETE_REQUEST.String():        DEV_HASH_DELETE_REQUEST,
		DEV_HASH_SAMPLE_REQUEST.String():        DEV_HASH_SAMPLE_REQUEST,
		DEV_LIVE_QUERY_CHANGE.String():          DEV_LIVE_QUERY_CHANGE,
		DEV_REREGISTER.String():                 DEV_REREGISTER,
		DEV_SENSOR_ACTION.String():              DEV_SENSOR_ACTION,
		DEV_SENSOR_CERTIFICATE_REQUEST.String(): DEV_SENSOR_CERTIFICATE_REQUEST,
		DEV_SENSOR_CONFIG_UPDATE.String():       DEV_SENSOR_CONFIG_UPDATE,
		DEV_SENSOR_KIT_UPDATE.String():          DEV_SENSOR_KIT_UPDATE,
		DEV_SENSOR_LOGS_REQUEST.String():        DEV_SENSOR_LOGS_REQUEST,
		DEV_SENSOR_TRIGGER_PERFMON.String():     DEV_SENSOR_TRIGGER_PERFMON,
	}

	AllDeviceChangeTypeArray = maps.Keys(TranslateDeviceChangeTypeString) //nolint:gochecknoglobals // we don't care
)
