// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

// Package types -- Contains various conversion structures.
package types

import (
	"database/sql/driver"
	"encoding/json"
	"errors"
	"fmt"
	"time"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"

	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
)

// DeploymentTypeToChangesMap -- A structure used by the Org model
//   - effectively "map[DeploymentType]map[OrgChangeType]timestamp"
type DeploymentTypeToChangesMap map[deploymenttype.DeploymentType]OrgChangeset

//goland:noinspection GoMixedReceiverTypes
func (dtc DeploymentTypeToChangesMap) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	var err error
	for deploymentType, changesMap := range dtc {
		err = encoder.AddObject(string(deploymentType), changesMap)
	}
	return err
}

// ----- Processing methods for DeploymentTypeToChangesMap --------------------

// Scan -- Convert incoming bytes into a DeploymentTypeToChangesMap structure
//
//goland:noinspection GoMixedReceiverTypes
func (dtc *DeploymentTypeToChangesMap) Scan(value interface{}) error {
	bytes, ok := value.([]byte)
	if !ok {
		return errors.New(fmt.Sprint("Failed to unmarshal JSONB value:", value))
	}
	return json.Unmarshal(bytes, dtc)
}

// Value -- return json value of DeploymentTypeToChangesMap, implement driver.Value interface
//
//goland:noinspection GoMixedReceiverTypes
func (dtc DeploymentTypeToChangesMap) Value() (driver.Value, error) {
	if len(dtc) == 0 {
		return nil, nil
	}
	rawBytes, err := json.Marshal(dtc)
	if err != nil {
		return nil, err
	}
	return json.RawMessage(rawBytes).MarshalJSON()
}

// NewDeploymentTypeChangesAtTime -- Creates a changes set with all org changes for a given time for a single deployment type
func NewDeploymentTypeChangesAtTime(deploymentType deploymenttype.DeploymentType, t time.Time) DeploymentTypeToChangesMap {
	deploymentChanges := make(DeploymentTypeToChangesMap)
	changes := NewChangesetAtTime(t)
	deploymentChanges[deploymentType] = changes
	return deploymentChanges
}

// NewDeploymentTypeToChangesMap -- generate a deployment type change set based on a supplied map (The inverse of toMap)
//   - Converts a 'map[string]map[string]int64' structure into a 'map[DeploymentType]map[OrgChangeType]timestamp' structure
func NewDeploymentTypeToChangesMap(deploymentTypeChanges map[string]map[string]map[string]int64, source zap.Field) DeploymentTypeToChangesMap {
	dtc := make(DeploymentTypeToChangesMap)

	for deploymentTypeName, changes := range deploymentTypeChanges {
		deploymentType := deploymenttype.FromString(deploymentTypeName)
		timestampMap := changes["timestampMap"]
		dtc[deploymentType] = NewOrgChangeSet(timestampMap, source)
		if !deploymentType.IsKnown() {
			logger.BaseLogger().Warn(fmt.Sprintf("Unknown DeploymentType '%s' seen for DeploymentTypeToChangesMap", deploymentTypeName),
				source)
		}
	}
	return dtc
}
