// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package repos

import (
	"context"
	"errors"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"go.uber.org/zap"
	"gorm.io/gorm"
	"gorm.io/gorm/clause"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/dbwrapper"

	"github.com/opentracing/opentracing-go"
)

// Repository -- Base repos interface used for all models
type Repository[T models.Model] interface {
	Get(ctx context.Context, model T) (err error)
	Update(ctx context.Context, model T) error
}

// GenericRepository -- Base repos structure
type GenericRepository[T models.Model] struct {
	db *dbwrapper.DbClient
}

func NewRepository[DBModel models.Model](db *dbwrapper.DbClient) GenericRepository[DBModel] {
	return GenericRepository[DBModel]{db: db}
}

// ----- Begin Repository methods --------------------------------------------------

// Get -- return a database entry by specific id for the given model
func (repo GenericRepository[T]) Get(ctx context.Context, model T) (err error) {
	sp := opentracing.SpanFromContext(ctx)
	if sp != nil {
		childSp := opentracing.StartSpan("DB QUERY", opentracing.ChildOf(sp.Context()))
		defer childSp.Finish()
	}
	err = repo.db.First(model).Error
	if err != nil && errors.Is(err, gorm.ErrRecordNotFound) {
		// We were trying to retrieve an object that we didn't know about yet so add it for future availability
		if err = repo.Update(ctx, model); err != nil {
			logger.RequestLoggerFromContext(ctx).Error("Failed to update the database with new model", zap.Error(err))
		} else {
			logger.RequestLoggerFromContext(ctx).Info("Added new object", zap.Object("Object", model))
		}
	}
	return
}

// Update -- Updates database entry with the latest values
func (repo GenericRepository[T]) Update(ctx context.Context, model T) (err error) {
	sp := opentracing.SpanFromContext(ctx)
	if sp != nil {
		childSp := opentracing.StartSpan("DB UPSERT", opentracing.ChildOf(sp.Context()))
		defer childSp.Finish()
	}
	// Update all columns, except primary keys, to new value on conflict
	logger.RequestLoggerFromContext(ctx).Debug("MODEL BEFORE UPDATE", zap.Object("model", model))
	err = repo.db.Model(model).Clauses(model.GetOnConflictClause(), clause.Returning{}).Create(model).Error
	logger.RequestLoggerFromContext(ctx).Debug("MODEL AFTER UPDATE", zap.Object("model", model))
	return err
}
