// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package dbwrapper

/* All credits to original authors of ems/device-auth */

import (
	"database/sql"
	"fmt"
	"sync/atomic"

	"gitlab.bit9.local/psc/ems/hinting/internal/config"
	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"

	gometrics "github.com/rcrowley/go-metrics"
)

// DBStat is an enum for the type of DB stat
type dbStat string

const (
	maxOpenConnections dbStat = "db.maxOpenConnections"
	openConnections    dbStat = "db.openConnections"
	inUse              dbStat = "db.inUse"
	idle               dbStat = "db.idle"
	waitCount          dbStat = "db.waitCount"
	waitDuration       dbStat = "db.waitDuration"
	maxIdleClosed      dbStat = "db.maxIdleClosed"
	maxLifetimeClosed  dbStat = "db.maxLifetimeClosed"
)

// dbStatGaugeMetric
type dbStatGaugeMetric struct {
	stat         dbStat
	db           SQLStatable
	cfg          *config.Config
	isReady      *int32
	isAlive      *int32
	history      []int64
	currentIndex int
}

func (m *dbStatGaugeMetric) Value() (val int64) {
	switch m.stat {
	case maxOpenConnections:
		val = int64(m.db.Stats().MaxOpenConnections)
	case openConnections:
		val = int64(m.db.Stats().OpenConnections)
	case inUse:
		val = int64(m.db.Stats().InUse)
	case waitCount:
		val = m.db.Stats().WaitCount
	case waitDuration:
		val = int64(m.db.Stats().WaitDuration)
	case maxIdleClosed:
		val = m.db.Stats().MaxIdleClosed
	case maxLifetimeClosed:
		val = m.db.Stats().MaxLifetimeClosed
	default:
		return 0
	}

	m.setHistory(val)
	m.ensureReadyAndAlive()
	return val
}

//goland:noinspection GoUnusedParameter
func (m *dbStatGaugeMetric) Update(_ int64) {
	panic("update not supported")
}

func (m *dbStatGaugeMetric) Snapshot() gometrics.Gauge {
	return gometrics.GaugeSnapshot(m.Value())
}

// setHistory advances the current index and adds the passed value to the history queue
func (m *dbStatGaugeMetric) setHistory(val int64) {
	if len(m.history) == 0 {
		return
	}
	m.currentIndex = (m.currentIndex + 1) % len(m.history)
	m.history[m.currentIndex] = val
}

// ensureReadyAndAlive sets the value of the healthy and live bool pointers based on the health of the metric
func (m *dbStatGaugeMetric) ensureReadyAndAlive() {
	if m.stat == waitCount {
		if len(m.history) < 2 {
			return
		}
		// if currently ready and history has increased => ready => false
		// if currently not ready and history constant => ready => true
		lastReady := atomic.LoadInt32(m.isReady) == 1
		lastAlive := atomic.LoadInt32(m.isAlive) == 1
		if (lastReady && !m.historyHasIncreased()) || (!lastReady && m.historyIsConstantOrDecreasing()) {
			atomic.StoreInt32(m.isReady, 1)
		} else {
			atomic.StoreInt32(m.isReady, 0)
		}
		if lastAlive && m.history[m.currentIndex] <= m.cfg.PostgresqlConfig.SQLWaitCountLimit {
			atomic.StoreInt32(m.isAlive, 1)
		} else {
			atomic.StoreInt32(m.isAlive, 0)
		}

		logger.BaseLogger().Debug(fmt.Sprintf("ensuring alive and ready from wait count history: %v, at index: %v, lastReady: %v, lastAlive: %v, currentReady: %v, currentAlive: %v", m.history, m.currentIndex, lastReady, lastAlive, m.isReady, m.isAlive))
	}
}

// historyHasIncreased returns true if from the oldest entry to the newest entry, any step is greater than the last
func (m *dbStatGaugeMetric) historyHasIncreased() bool {
	historyLength := len(m.history)
	// (currentIndex + 1) % historyLength = the oldest value in history
	last := m.history[(m.currentIndex+1)%historyLength]
	// compare the next (historyLength - 1) history values to see if values are increasing over time
	for i := 0; i < historyLength-1; i++ {
		// the earliest value is +1, so our base comparison is +2
		next := m.history[(m.currentIndex+2+i)%historyLength]
		// value increased over one iteration, history has increased
		if next > last {
			return true
		}
		last = next
	}
	return false
}

// historyIsConstantOrDecreasing returns true if the newest value is less than or equal to the oldest value
func (m *dbStatGaugeMetric) historyIsConstantOrDecreasing() bool {
	// if current history value is less than or equal to the oldest history value, we are constant or decreasing
	return m.history[m.currentIndex] <= m.history[(m.currentIndex+1)%len(m.history)]
}

// RegisterDBStatMetrics registers Wavefront metrics for each DB stat
func RegisterDBStatMetrics(db SQLStatable, service metrics.Service, isReady, isAlive *int32, cfg *config.Config) {
	dbStats := []dbStat{
		maxOpenConnections,
		openConnections,
		inUse,
		idle,
		waitCount,
		waitDuration,
		maxIdleClosed,
		maxLifetimeClosed,
	}

	for _, stat := range dbStats {
		if err := service.RegisterMetric(string(stat),
			&dbStatGaugeMetric{
				db: db, stat: stat, cfg: cfg, isReady: isReady, isAlive: isAlive,
				history: make([]int64, cfg.PostgresqlConfig.SQLStatTrackingDepth),
			}, cfg.BaseDimensions); err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("error registering db metric: %v", stat))
		} else {
			logger.BaseLogger().Debug(fmt.Sprintf("Registered db metric: %v", stat))
		}
	}
}

// SQLStatable is an interface to get stats from sql
type SQLStatable interface {
	Stats() sql.DBStats
}
