// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package redisclient

import (
	"context"
	"errors"
	"fmt"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"

	"gitlab.bit9.local/psc/ems/hinting/internal/config"
)

var ErrRedisNil = errors.New("redis nil")

type RedisClient interface {
	Get(ctx context.Context, cacheKey string) ([]byte, error)
	Set(ctx context.Context, cacheKey string, data []byte) error
	Del(ctx context.Context, cacheKey string) error
	Ping(ctx context.Context) error
	Close() error
}

func getRedisClientAndPing(cfg *config.Config) (client RedisClient, err error) {
	client, err = getClientbyType(cfg)
	if err != nil {
		return nil, fmt.Errorf("failure to create redis client %w", err)
	}
	if err = waitForRedisClient(100, client); err != nil {
		return nil, fmt.Errorf("redis client did not respond: %w", err)
	} else {
		logger.BaseLogger().Debug("Successfully connected to redis.")
	}
	return
}

func getClientbyType(cfg *config.Config) (RedisClient, error) {
	switch cfg.NoSQLClientType {
	case "goredis":
		return newGoRedisClient(cfg)
	default:
		return newRueidisClientFromConfig(cfg)
	}
}

func waitForRedisClient(retries int, client RedisClient) error {
	var err error = nil
	for i := 0; i < retries; i++ {
		err = client.Ping(context.Background())
		if err != nil {
			time.Sleep(100 * time.Millisecond)
		} else {
			return nil
		}
	}
	return err
}
