package redisclient

import (
	"context"
	"crypto/tls"
	"strings"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/config"

	"github.com/rueian/rueidis"
)

type rueidisClient struct {
	client  rueidis.Client
	Cluster string
	Cfg     *config.Config
}

func (r *rueidisClient) Get(ctx context.Context, cacheKey string) ([]byte, error) {
	result := r.client.DoCache(ctx, r.client.B().Get().Key(cacheKey).Cache(), 15*time.Minute)
	stringData, err := result.ToString()
	if err != nil {
		return nil, wrapRedisNil(err)
	} else {
		return []byte(stringData), nil
	}
}

func (r *rueidisClient) Set(ctx context.Context, cacheKey string, data []byte) error {
	return r.client.Do(ctx, r.client.B().Set().Key(cacheKey).Value(string(data)).ExSeconds(int64(r.Cfg.NoSQLTTL.Seconds())).Build()).NonRedisError()
}

func (r *rueidisClient) Del(ctx context.Context, cacheKey string) error {
	return r.client.Do(ctx, r.client.B().Unlink().Key(cacheKey).Build()).NonRedisError()
}

func (r *rueidisClient) Close() error {
	r.client.Close()
	return nil
}

func (r *rueidisClient) Ping(ctx context.Context) error {
	return r.client.Do(ctx, r.client.B().Ping().Build()).Error()
}

func newRueidisClientFromConfig(cfg *config.Config) (RedisClient, error) {
	var underlyingClient rueidis.Client
	var err error
	if cfg.IsLocalEnv() {
		underlyingClient, err = newLocalRedisClientFromConfig(cfg)
	} else {
		underlyingClient, err = newRedisClusterClientFromConfig(cfg)
	}
	if err != nil {
		return nil, err
	}
	return &rueidisClient{Cfg: cfg, client: underlyingClient}, nil
}

func newLocalRedisClientFromConfig(cfg *config.Config) (rueidis.Client, error) {
	return rueidis.NewClient(rueidis.ClientOption{
		Password:         cfg.NoSQLAuthToken,
		ConnWriteTimeout: cfg.NoSQLWriteTimeout,
		InitAddress:      []string{cfg.NoSQLConnection},
		DisableCache:     true,
	})
}

func newRedisClusterClientFromConfig(cfg *config.Config) (rueidis.Client, error) {
	return newRedisClusterClient(cfg.NoSQLTLS, cfg.NoSQLAuthToken, cfg.NoSQLWriteTimeout, cfg.NoSQLConnection)
}

func newRedisClusterClient(useTls bool, token string, timeout time.Duration, addresses string) (rueidis.Client, error) {
	var tlsConfig *tls.Config
	if useTls {
		tlsConfig = &tls.Config{MinVersion: tls.VersionTLS12}
	}
	return rueidis.NewClient(rueidis.ClientOption{
		Password:         token,
		ConnWriteTimeout: timeout,
		InitAddress:      strings.Split(addresses, ","),
		TLSConfig:        tlsConfig,
	})
}

func wrapRedisNil(err error) error {
	if rueidis.IsRedisNil(err) {
		return ErrRedisNil
	}
	return err
}
